r"""
Algebra modules
"""
#*****************************************************************************
#  Copyright (C) 2005      David Kohel <kohel@maths.usyd.edu>
#                          William Stein <wstein@math.ucsd.edu>
#                2008-2009 Nicolas M. Thiery <nthiery at users.sf.net>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#******************************************************************************

from .category_types import Category_module
from .modules import Modules

class AlgebraModules(Category_module):
    """
    The category of modules over a fixed algebra `A`.

    EXAMPLES::

        sage: AlgebraModules(QQ['a'])
        Category of algebra modules over Univariate Polynomial Ring in a over Rational Field
        sage: AlgebraModules(QQ['a']).super_categories()
        [Category of modules over Univariate Polynomial Ring in a over Rational Field]

    Note: as of now, `A` is required to be commutative, ensuring that
    the categories of left and right modules are isomorphic. Feedback
    and use cases for potential generalizations to the non commutative
    case are welcome.

    """
    def __init__(self, A):
        """
        EXAMPLES::

            sage: AlgebraModules(QQ['a'])
            Category of algebra modules over Univariate Polynomial Ring in a over Rational Field
            sage: AlgebraModules(QQ['a,b']) # todo: not implemented (QQ['a,b'] should be in Algebras(QQ))
            sage: AlgebraModules(FreeAlgebra(QQ,2,'a,b'))
            Traceback (most recent call last):
            ...
            TypeError: A (=Free Algebra on 2 generators (a, b) over Rational Field) must be a commutative algebra
            sage: AlgebraModules(QQ)
            Traceback (most recent call last):
            ...
            TypeError: A (=Rational Field) must be a commutative algebra

        TESTS::

            sage: TestSuite(AlgebraModules(QQ['a'])).run()
        """
        from sage.categories.commutative_algebras import CommutativeAlgebras
        if not hasattr(A, "base_ring") or A not in CommutativeAlgebras(A.base_ring()):
            raise TypeError("A (=%s) must be a commutative algebra" % A)
        Category_module.__init__(self, A)

    @classmethod
    def an_instance(cls):
        """
        Returns an instance of this class

        EXAMPLES::

            sage: AlgebraModules.an_instance()
            Category of algebra modules over Univariate Polynomial Ring in x over Rational Field
        """
        from sage.rings.rational_field import QQ
        return cls(QQ['x'])

    def algebra(self):
        """
        EXAMPLES::

            sage: AlgebraModules(QQ['x']).algebra()
            Univariate Polynomial Ring in x over Rational Field
        """
        return self.base_ring()

    def super_categories(self):
        """
        EXAMPLES::

            sage: AlgebraModules(QQ['x']).super_categories()
            [Category of modules over Univariate Polynomial Ring in x over Rational Field]
        """
        R = self.algebra()
        return [Modules(R)]



r"""
Graded algebras with basis
"""
#*****************************************************************************
#  Copyright (C) 2008      Teresa Gomez-Diaz (CNRS) <Teresa.Gomez-Diaz@univ-mlv.fr>
#                2008-2011 Nicolas M. Thiery <nthiery at users.sf.net>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#******************************************************************************

from sage.categories.graded_modules import GradedModulesCategory
from sage.categories.signed_tensor import SignedTensorProductsCategory, tensor_signed
from sage.misc.cachefunc import cached_method

class GradedAlgebrasWithBasis(GradedModulesCategory):
    """
    The category of graded algebras with a distinguished basis

    EXAMPLES::

        sage: C = GradedAlgebrasWithBasis(ZZ); C
        Category of graded algebras with basis over Integer Ring
        sage: sorted(C.super_categories(), key=str)
        [Category of filtered algebras with basis over Integer Ring,
         Category of graded algebras over Integer Ring,
         Category of graded modules with basis over Integer Ring]

    TESTS::

        sage: TestSuite(C).run()
    """
    class ParentMethods:
        # This needs to be copied in GradedAlgebras because we need to have
        #   FilteredAlgebrasWithBasis as an extra super category
        def graded_algebra(self):
            """
            Return the associated graded algebra to ``self``.

            This is ``self``, because ``self`` is already graded.
            See :meth:`~sage.categories.filtered_algebras_with_basis.FilteredAlgebrasWithBasis.graded_algebra`
            for the general behavior of this method, and see
            :class:`~sage.algebras.associated_graded.AssociatedGradedAlgebra`
            for the definition and properties of associated graded
            algebras.

            EXAMPLES::

                sage: m = SymmetricFunctions(QQ).m()
                sage: m.graded_algebra() is m
                True

            TESTS:

            Let us check that the three methods
            :meth:`to_graded_conversion`, :meth:`from_graded_conversion`
            and :meth:`projection` (which form the interface of the
            associated graded algebra) work correctly here::

                sage: to_gr = m.to_graded_conversion()
                sage: from_gr = m.from_graded_conversion()
                sage: m[2] == to_gr(m[2]) == from_gr(m[2])
                True
                sage: u = 3*m[1] - (1/2)*m[3]
                sage: u == to_gr(u) == from_gr(u)
                True
                sage: m.zero() == to_gr(m.zero()) == from_gr(m.zero())
                True
                sage: p2 = m.projection(2)
                sage: p2(m[2] - 4*m[1,1] + 3*m[1] - 2*m[[]])
                -4*m[1, 1] + m[2]
                sage: p2(4*m[1])
                0
                sage: p2(m.zero()) == m.zero()
                True
            """
            return self

        # .. TODO::
        # Possibly override ``to_graded_conversion`` and
        # ``from_graded_conversion`` with identity morphisms?
        # I have to admit I don't know the right way to construct
        # identity morphisms other than using the identity matrix.
        # Also, ``projection`` could be overridden by, well, a
        # projection.

        def free_graded_module(self, generator_degrees, names=None):
            """
            Create a finitely generated free graded module over ``self``

            INPUT:

            - ``generator_degrees`` -- tuple of integers defining the
              number of generators of the module and their degrees

            - ``names`` -- (optional) the names of the generators. If
              ``names`` is a comma-separated string like ``'a, b,
              c'``, then those will be the names. Otherwise, for
              example if ``names`` is ``abc``, then the names will be
              ``abc[d,i]``.

            By default, if all generators are in distinct degrees,
            then the ``names`` of the generators will have the form
            ``g[d]`` where ``d`` is the degree of the generator. If
            the degrees are not distinct, then the generators will be
            called ``g[d,i]`` where ``d`` is the degree and ``i`` is
            its index in the list of generators in that degree.

            See :mod:`sage.modules.fp_graded.free_module` for more
            examples and details.

            EXAMPLES::

                sage: Q = QuadraticForm(QQ, 3, [1,2,3,4,5,6])
                sage: Cl = CliffordAlgebra(Q)
                sage: M = Cl.free_graded_module((0, 2, 3))
                sage: M.gens()
                (g[0], g[2], g[3])
                sage: N.<xy, z> = Cl.free_graded_module((1, 2))
                sage: N.generators()
                (xy, z)
            """
            try:
                return self._free_graded_module_class(self, generator_degrees, names=names)
            except AttributeError:
                from sage.modules.fp_graded.free_module import FreeGradedModule
                return FreeGradedModule(self, generator_degrees, names=names)

        def formal_series_ring(self):
            r"""
            Return the completion of all formal linear combinations of
            ``self`` with finite linear combinations in each homogeneous
            degree (computed lazily).

            EXAMPLES::

                sage: NCSF = NonCommutativeSymmetricFunctions(QQ)
                sage: S = NCSF.Complete()
                sage: L = S.formal_series_ring()
                sage: L
                Lazy completion of Non-Commutative Symmetric Functions over
                 the Rational Field in the Complete basis
            """
            from sage.rings.lazy_series_ring import LazyCompletionGradedAlgebra
            return LazyCompletionGradedAlgebra(self)

    class ElementMethods:
        pass

    class SignedTensorProducts(SignedTensorProductsCategory):
        """
        The category of algebras with basis constructed by signed tensor
        product of algebras with basis.
        """
        @cached_method
        def extra_super_categories(self):
            """
            EXAMPLES::

                sage: Cat = AlgebrasWithBasis(QQ).Graded()
                sage: Cat.SignedTensorProducts().extra_super_categories()
                [Category of graded algebras with basis over Rational Field]
                sage: Cat.SignedTensorProducts().super_categories()
                [Category of graded algebras with basis over Rational Field,
                 Category of signed tensor products of graded algebras over Rational Field]
            """
            return [self.base_category()]

        class ParentMethods:
            """
            Implements operations on tensor products of super algebras
            with basis.
            """
            @cached_method
            def one_basis(self):
                """
                Return the index of the one of this signed tensor product of
                algebras, as per ``AlgebrasWithBasis.ParentMethods.one_basis``.

                It is the tuple whose operands are the indices of the
                ones of the operands, as returned by their
                :meth:`.one_basis` methods.

                EXAMPLES::

                    sage: A.<x,y> = ExteriorAlgebra(QQ)
                    sage: A.one_basis()
                    0
                    sage: B = tensor((A, A, A))
                    sage: B.one_basis()
                    (0, 0, 0)
                    sage: B.one()
                    1 # 1 # 1
                """
                # FIXME: this method should be conditionally defined,
                # so that B.one_basis returns NotImplemented if not
                # all modules provide one_basis
                if all(hasattr(module, "one_basis") for module in self._sets):
                    return tuple(module.one_basis() for module in self._sets)
                else:
                    raise NotImplementedError

            def product_on_basis(self, t0, t1):
                """
                The product of the algebra on the basis, as per
                ``AlgebrasWithBasis.ParentMethods.product_on_basis``.

                EXAMPLES:

                Test the sign in the super tensor product::

                    sage: A = SteenrodAlgebra(3)
                    sage: x = A.Q(0)
                    sage: y = x.coproduct()
                    sage: y^2
                    0

                TODO: optimize this implementation!
                """
                basic = tensor_signed((module.monomial(x0) * module.monomial(x1)
                                      for (module, x0, x1) in zip(self._sets, t0, t1)))
                n = len(self._sets)
                parity0 = [self._sets[idx].degree_on_basis(x0)
                           for (idx, x0) in enumerate(t0)]
                parity1 = [self._sets[idx].degree_on_basis(x1)
                           for (idx, x1) in enumerate(t1)]
                parity = sum(parity0[i] * parity1[j]
                             for j in range(n) for i in range(j+1,n))
                return (-1)**parity * basic



r"""
Supercommutative Algebras
"""
#*****************************************************************************
#  Copyright (C) 2019 Travis Scrimshaw <tcscrims at gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from sage.categories.category_with_axiom import CategoryWithAxiom_over_base_ring
from sage.categories.super_algebras import SuperAlgebras
from sage.categories.signed_tensor import SignedTensorProductsCategory
from sage.misc.cachefunc import cached_method

class SupercommutativeAlgebras(CategoryWithAxiom_over_base_ring):
    r"""
    The category of supercommutative algebras.

    An `R`-*supercommutative algebra* is an `R`-super algebra
    `A = A_0 \oplus A_1` endowed with an `R`-super algebra structure
    satisfying:

    .. MATH::

        x_0 x'_0 = x'_0 x_0, \qquad
        x_1 x'_1 = -x'_1 x_1, \qquad
        x_0 x_1 = x_1 x_0,

    for all `x_0, x'_0 \in A_0` and `x_1, x'_1 \in A_1`.

    EXAMPLES::

        sage: Algebras(ZZ).Supercommutative()
        Category of supercommutative algebras over Integer Ring

    TESTS::

        sage: TestSuite(Algebras(ZZ).Supercommutative()).run()
    """
    _base_category_class_and_axiom = (SuperAlgebras, "Supercommutative")

    class SignedTensorProducts(SignedTensorProductsCategory):
        @cached_method
        def extra_super_categories(self):
            """
            Return the extra super categories of ``self``.

            A signed tensor product of supercommutative algebras is a
            supercommutative algebra.

            EXAMPLES::

                sage: C = Algebras(ZZ).Supercommutative().SignedTensorProducts()
                sage: C.extra_super_categories()
                [Category of supercommutative algebras over Integer Ring]
            """
            return [self.base_category()]

    class WithBasis(CategoryWithAxiom_over_base_ring):
        class ParentMethods:
            def _test_supercommutativity(self, **options):
                r"""
                Test supercommutativity for (not necessarily all) elements
                of this supercommutative algebra.

                INPUT:

                - ``options`` -- any keyword arguments accepted by :meth:`_tester`

                EXAMPLES:

                By default, this method tests only the elements returned by
                ``self.some_elements()``::

                    sage: E.<x,y,z> = ExteriorAlgebra(QQ)
                    sage: E._test_supercommutativity()

                However, the elements tested can be customized with the
                ``elements`` keyword argument, but the elements must be
                homogeneous::

                    sage: E._test_supercommutativity(elements=[x+y, x*y-3*y*z, x*y*z])
                    sage: E._test_supercommutativity(elements=[x+x*y])
                    Traceback (most recent call last):
                    ...
                    ValueError: element is not homogeneous

                See the documentation for :class:`TestSuite` for more information.
                """
                elements = options.pop("elements", self.basis())
                tester = self._tester(**options)
                from sage.misc.misc import some_tuples
                for x,y in some_tuples(elements, 2, tester._max_runs):
                    tester.assertEqual((x * y),
                                       (-1)**(x.is_even_odd() * y.is_even_odd())
                                       * (y * x))



r"""
Implementing a new parent: a tutorial


The easiest approach for implementing a new parent is to start from a
close example in sage.categories.examples. Here, we will get through
the process of implementing a new finite semigroup, taking as starting
point the provided example::

    sage: S = FiniteSemigroups().example()
    sage: S
    An example of a finite semigroup: the left regular band generated by ('a', 'b', 'c', 'd')

You may lookup the implementation of this example with::

    sage: S??                               # not tested

Or by browsing the source code of
:class:`sage.categories.examples.finite_semigroups.LeftRegularBand`.

Copy-paste this code into, say, a cell of the notebook, and replace
every occurrence of ``FiniteSemigroups().example(...)`` in the
documentation by ``LeftRegularBand``. This will be equivalent to::

    sage: from sage.categories.examples.finite_semigroups import LeftRegularBand

Now, try::

    sage: S = LeftRegularBand(); S
    An example of a finite semigroup: the left regular band generated by ('a', 'b', 'c', 'd')

and play around with the examples in the documentation of ``S`` and of
:class:`FiniteSemigroups`.

Rename the class to ``ShiftSemigroup``, and modify the product to
implement the semigroup generated by the given alphabet such that `au
= u` for any `u` of length `3`.

Use ``TestSuite`` to test the newly implemented semigroup; draw its
Cayley graph.

Add another option to the constructor to generalize the construction
to any u of length `k`.

Lookup the Sloane for the sequence of the sizes of those semigroups.


Now implement the commutative monoid of subsets of `\{1,\dots,n\}`
endowed with union as product. What is its category? What are the
extra functionalities available there? Implement iteration and
cardinality.

.. TODO::

   The tutorial should explain there how to reuse the enumerated
   set of subsets, and endow it with more structure.

"""



r"""
Finite Crystals
"""
#*****************************************************************************
#  Copyright (C) 2010    Anne Schilling <anne at math.ucdavis.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#******************************************************************************

from sage.misc.cachefunc import cached_method
from sage.categories.category_with_axiom import CategoryWithAxiom
from sage.categories.finite_enumerated_sets import FiniteEnumeratedSets
from sage.categories.tensor import TensorProductsCategory

class FiniteCrystals(CategoryWithAxiom):
    """
    The category of finite crystals.

    EXAMPLES::

        sage: C = FiniteCrystals()
        sage: C
        Category of finite crystals
        sage: C.super_categories()
        [Category of crystals, Category of finite enumerated sets]
        sage: C.example()
        Highest weight crystal of type A_3 of highest weight omega_1

    TESTS::

        sage: TestSuite(C).run()
        sage: B = FiniteCrystals().example()
        sage: TestSuite(B).run(verbose = True)
        running ._test_an_element() . . . pass
        running ._test_cardinality() . . . pass
        running ._test_category() . . . pass
        running ._test_construction() . . . pass
        running ._test_elements() . . .
          Running the test suite of self.an_element()
          running ._test_category() . . . pass
          running ._test_eq() . . . pass
          running ._test_new() . . . pass
          running ._test_not_implemented_methods() . . . pass
          running ._test_pickling() . . . pass
          running ._test_stembridge_local_axioms() . . . pass
          pass
        running ._test_elements_eq_reflexive() . . . pass
        running ._test_elements_eq_symmetric() . . . pass
        running ._test_elements_eq_transitive() . . . pass
        running ._test_elements_neq() . . . pass
        running ._test_enumerated_set_contains() . . . pass
        running ._test_enumerated_set_iter_cardinality() . . . pass
        running ._test_enumerated_set_iter_list() . . . pass
        running ._test_eq() . . . pass
        running ._test_fast_iter() . . . pass
        running ._test_new() . . . pass
        running ._test_not_implemented_methods() . . . pass
        running ._test_pickling() . . . pass
        running ._test_some_elements() . . . pass
        running ._test_stembridge_local_axioms() . . . pass
    """

    @cached_method
    def extra_super_categories(self):
        r"""
        EXAMPLES::

            sage: FiniteCrystals().extra_super_categories()
            [Category of finite enumerated sets]
        """
        return [FiniteEnumeratedSets()]

    def example(self, n=3):
        """
        Returns an example of highest weight crystals, as per
        :meth:`Category.example`.

        EXAMPLES::

            sage: B = FiniteCrystals().example(); B
            Highest weight crystal of type A_3 of highest weight omega_1
        """
        from sage.categories.crystals import Crystals
        return Crystals().example(n)

    class TensorProducts(TensorProductsCategory):
        """
        The category of finite crystals constructed by tensor
        product of finite crystals.
        """
        @cached_method
        def extra_super_categories(self):
            """
            EXAMPLES::

                sage: FiniteCrystals().TensorProducts().extra_super_categories()
                [Category of finite crystals]
            """
            return [self.base_category()]



# sage.doctest: optional - sage.symbolic
r"""
Vector Bundles
"""
#*****************************************************************************
#  Copyright (C) 2019 Michael Jung <micjung at uni-potsdam.de>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#******************************************************************************

from sage.categories.category_types import Category_over_base_ring
from sage.categories.category_with_axiom import CategoryWithAxiom_over_base_ring
from sage.misc.cachefunc import cached_method
from sage.categories.sets_cat import Sets
from sage.categories.fields import Fields

class VectorBundles(Category_over_base_ring):
    r"""
    The category of vector bundles over any base space and base field.

    .. SEEALSO:: :class:`~sage.manifolds.vector_bundle.TopologicalVectorBundle`

    EXAMPLES::

        sage: M = Manifold(2, 'M', structure='top')
        sage: from sage.categories.vector_bundles import VectorBundles
        sage: C = VectorBundles(M, RR); C
        Category of vector bundles over Real Field with 53 bits of precision
         with base space 2-dimensional topological manifold M
        sage: C.super_categories()
        [Category of topological spaces]

    TESTS::

        sage: TestSuite(C).run(skip="_test_category_over_bases")

    """
    def __init__(self, base_space, base_field, name=None):
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: M = Manifold(2, 'M')
            sage: from sage.categories.vector_bundles import VectorBundles
            sage: C = VectorBundles(M, RR)
            sage: TestSuite(C).run(skip="_test_category_over_bases")

        """
        if base_field not in Fields().Topological():
            raise ValueError("base field must be a topological field")
        self._base_space = base_space
        Category_over_base_ring.__init__(self, base_field, name)

    @cached_method
    def super_categories(self):
        r"""
        EXAMPLES::

            sage: M = Manifold(2, 'M')
            sage: from sage.categories.vector_bundles import VectorBundles
            sage: VectorBundles(M, RR).super_categories()
            [Category of topological spaces]

        """
        return [Sets().Topological()]

    def base_space(self):
        r"""
        Return the base space of this category.

        EXAMPLES::

            sage: M = Manifold(2, 'M', structure='top')
            sage: from sage.categories.vector_bundles import VectorBundles
            sage: VectorBundles(M, RR).base_space()
            2-dimensional topological manifold M

        """
        return self._base_space

    def _repr_object_names(self):
        r"""
        Return the name of the objects of this category.

        .. SEEALSO:: :meth:`Category._repr_object_names`

        EXAMPLES::

            sage: M = Manifold(2, 'M')
            sage: from sage.categories.vector_bundles import VectorBundles
            sage: VectorBundles(M, RR)._repr_object_names()
            'vector bundles over Real Field with 53 bits of precision with base
             space 2-dimensional differentiable manifold M'

        """
        base_space = self._base_space
        return Category_over_base_ring._repr_object_names(self) + \
               " with base space %s"%base_space

    class SubcategoryMethods:
        @cached_method
        def Differentiable(self):
            r"""
            Return the subcategory of the differentiable objects
            of ``self``.

            EXAMPLES::

                sage: M = Manifold(2, 'M')
                sage: from sage.categories.vector_bundles import VectorBundles
                sage: VectorBundles(M, RR).Differentiable()
                Category of differentiable vector bundles over Real Field with
                 53 bits of precision with base space 2-dimensional
                 differentiable manifold M

            TESTS::

                sage: TestSuite(VectorBundles(M, RR).Differentiable()).run()
                sage: VectorBundles(M, RR).Differentiable.__module__
                'sage.categories.vector_bundles'

            """
            return self._with_axiom('Differentiable')

        @cached_method
        def Smooth(self):
            """
            Return the subcategory of the smooth objects of ``self``.

            EXAMPLES::

                sage: M = Manifold(2, 'M')
                sage: from sage.categories.vector_bundles import VectorBundles
                sage: VectorBundles(M, RR).Smooth()
                Category of smooth vector bundles over Real Field with 53 bits
                 of precision with base space 2-dimensional differentiable
                 manifold M

            TESTS::

                sage: TestSuite(VectorBundles(M, RR).Smooth()).run()
                sage: VectorBundles(M, RR).Smooth.__module__
                'sage.categories.vector_bundles'
            """
            return self._with_axiom('Smooth')

    class Differentiable(CategoryWithAxiom_over_base_ring):
        """
        The category of differentiable vector bundles.

        A differentiable vector bundle is a differentiable manifold with
        differentiable surjective projection on a differentiable base space.
        """

    class Smooth(CategoryWithAxiom_over_base_ring):
        """
        The category of smooth vector bundles.

        A smooth vector bundle is a smooth manifold with
        smooth surjective projection on a smooth base space.
        """



r"""
Finitely generated magmas
"""
#*****************************************************************************
#  Copyright (C) 2014 Nicolas M. Thiery <nthiery at users.sf.net>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#******************************************************************************

from sage.misc.abstract_method import abstract_method
from sage.categories.category_with_axiom import CategoryWithAxiom
from sage.categories.magmas import Magmas

class FinitelyGeneratedMagmas(CategoryWithAxiom):
    r"""
    The category of finitely generated (multiplicative) magmas.

    See :meth:`Magmas.SubcategoryMethods.FinitelyGeneratedAsMagma` for
    details.

    EXAMPLES::

        sage: C = Magmas().FinitelyGeneratedAsMagma(); C
        Category of finitely generated magmas
        sage: C.super_categories()
        [Category of magmas]
        sage: sorted(C.axioms())
        ['FinitelyGeneratedAsMagma']

    TESTS::

        sage: TestSuite(C).run()
    """

    _base_category_class_and_axiom = (Magmas, "FinitelyGeneratedAsMagma")

    class ParentMethods:

        @abstract_method
        def magma_generators(self):
            """
            Return distinguished magma generators for ``self``.

            OUTPUT: a finite family

            This method should be implemented by all
            :class:`finitely generated magmas <FinitelyGeneratedMagmas>`.

            EXAMPLES::

                sage: S = FiniteSemigroups().example()
                sage: S.magma_generators()
                Family ('a', 'b', 'c', 'd')
            """



r"""
Domains
"""
#*****************************************************************************
#  Copyright (C) 2008 Teresa Gomez-Diaz (CNRS) <Teresa.Gomez-Diaz@univ-mlv.fr>
#                2012 Nicolas M. Thiery <nthiery at users.sf.net>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#******************************************************************************

from sage.misc.lazy_import import LazyImport
from sage.categories.category_with_axiom import CategoryWithAxiom
from sage.categories.rings import Rings

class Domains(CategoryWithAxiom):
    """
    The category of domains

    A domain (or non-commutative integral domain), is a ring, not
    necessarily commutative, with no nonzero zero divisors.

    EXAMPLES::

        sage: C = Domains(); C
        Category of domains
        sage: C.super_categories()
        [Category of rings]
        sage: C is Rings().NoZeroDivisors()
        True

    TESTS::

        sage: TestSuite(C).run()
    """

    _base_category_class_and_axiom = (Rings, "NoZeroDivisors")

    def super_categories(self):
        """
        EXAMPLES::

            sage: Domains().super_categories()
            [Category of rings]
        """
        return [Rings()]

    Commutative = LazyImport('sage.categories.integral_domains', 'IntegralDomains', at_startup=True)

    class ParentMethods:
        def _test_zero_divisors(self, **options):
            """
            Check to see that there are no zero divisors.

            .. NOTE::

                In rings whose elements can not be represented exactly, there
                may be zero divisors in practice, even though these rings do
                not have them in theory. For such inexact rings, these tests
                are not performed::

                    sage: R = ZpFM(5); R
                    5-adic Ring of fixed modulus 5^20
                    sage: R.is_exact()
                    False
                    sage: a = R(5^19)
                    sage: a.is_zero()
                    False
                    sage: (a*a).is_zero()
                    True
                    sage: R._test_zero_divisors()

            EXAMPLES::

                sage: ZZ._test_zero_divisors()
                sage: ZpFM(5)._test_zero_divisors()

            """
            if not self.is_exact():
                return # Can't check on inexact rings

            tester = self._tester(**options)

            # Filter out zero
            S = [s for s in tester.some_elements() if not s.is_zero()]

            from sage.misc.misc import some_tuples
            for a,b in some_tuples(S, 2, tester._max_runs):
                p = a * b
                tester.assertFalse(p.is_zero())

    class ElementMethods:
        pass



r"""
Drinfeld modules over a base

This module provides the class
:class:`sage.category.drinfeld_modules.DrinfeldModules`.

AUTHORS:

- Antoine Leudière (2022-04)
- Xavier Caruso (2022-06)
"""

# *****************************************************************************
#   Copyright (C) 2022      Xavier Caruso <xavier.caruso@normalesup.org>
#                           Antoine Leudière <antoine.leudiere@inria.fr>
#
#   Distributed under the terms of the GNU General Public License (GPL)
#                   http://www.gnu.org/licenses/
# ******************************************************************************

from sage.categories.objects import Objects
from sage.categories.category_types import Category_over_base_ring
from sage.categories.homsets import Homsets
from sage.misc.functional import log
from sage.misc.latex import latex
from sage.rings.integer import Integer
from sage.rings.polynomial.ore_polynomial_ring import OrePolynomialRing
from sage.rings.polynomial.polynomial_ring import PolynomialRing_general
from sage.rings.ring_extension import RingExtension_generic


class DrinfeldModules(Category_over_base_ring):
    r"""
    This class implements the category of Drinfeld
    `\mathbb{F}_q[T]`-modules on a given base field.

    Let `\mathbb{F}_q[T]` be a polynomial ring with coefficients in a
    finite field `\mathbb{F}_q` and let `K` be a field. Fix a ring
    morphism `\gamma: \mathbb{F}_q[T] \to K`; we say that `K` is an
    `\mathbb{F}_q[T]`*-field*. Let `K\{\tau\}` be the ring of Ore
    polynomials with coefficients in `K`, whose multiplication is given
    by the rule `\tau \lambda = \lambda^q \tau` for any `\lambda \in K`.

    The extension `K`/`\mathbb{F}_q[T]` (represented as an instance of
    the class :class:`sage.rings.ring_extension.RingExtension`) is the
    *base field* of the category; its defining morphism `\gamma` is
    called the *base morphism*.

    The monic polynomial that generates the kernel of `\gamma` is called
    the `\mathbb{F}_q[T]`-*characteristic*, or *function-field
    characteristic*, of the base field. We say that `\mathbb{F}_q[T]` is
    the *function ring* of the category; `K\{\tau\}` is the *Ore
    polynomial ring*. The constant coefficient of the category is the
    image of `T` under the base morphism.

    .. RUBRIC:: Construction

    Generally, Drinfeld modules objects are created before their
    category, and the category is retrieved as an attribute of the
    Drinfeld module::

        sage: Fq = GF(11)
        sage: A.<T> = Fq[]
        sage: K.<z> = Fq.extension(4)
        sage: p_root = z^3 + 7*z^2 + 6*z + 10
        sage: phi = DrinfeldModule(A, [p_root, 0, 0, 1])
        sage: C = phi.category()
        sage: C
        Category of Drinfeld modules over Finite Field in z of size 11^4 over its base

    The output tells the user that the category is only defined by its
    base.

    .. RUBRIC:: Properties of the category

    The base field is retrieved using the method :meth:`base`.

        sage: C.base()
        Finite Field in z of size 11^4 over its base

    Equivalently, one can use :meth:`base_morphism` to retrieve the base
    morphism::

        sage: C.base_morphism()
        Ring morphism:
          From: Univariate Polynomial Ring in T over Finite Field of size 11
          To:   Finite Field in z of size 11^4 over its base
          Defn: T |--> z^3 + 7*z^2 + 6*z + 10

    The so-called constant coefficient --- which is the same for all
    Drinfeld modules in the category --- is simply the image of `T` by
    the base morphism::

        sage: C.constant_coefficient()
        z^3 + 7*z^2 + 6*z + 10
        sage: C.base_morphism()(T) == C.constant_coefficient()
        True

    Similarly, the function ring-characteristic of the category is
    either `0` or the unique monic polynomial in `\mathbb{F}_q[T]` that
    generates the kernel of the base::

        sage: C.characteristic()
        T^2 + 7*T + 2
        sage: C.base_morphism()(C.characteristic())
        0

    The base field, base morphism, function ring and Ore polynomial ring
    are the same for the category and its objects::

        sage: C.base() is phi.base()
        True
        sage: C.base_morphism() is phi.base_morphism()
        True

        sage: C.function_ring()
        Univariate Polynomial Ring in T over Finite Field of size 11
        sage: C.function_ring() is phi.function_ring()
        True

        sage: C.ore_polring()
        Ore Polynomial Ring in t over Finite Field in z of size 11^4 over its base twisted by Frob
        sage: C.ore_polring() is phi.ore_polring()
        True


    .. RUBRIC:: Creating Drinfeld module objects from the category

    Calling :meth:`object` with an Ore polynomial creates a Drinfeld module
    object in the category whose generator is the input::

        sage: psi = C.object([p_root, 1])
        sage: psi
        Drinfeld module defined by T |--> t + z^3 + 7*z^2 + 6*z + 10
        sage: psi.category() is C
        True

    Of course, the constant coefficient of the input must be the same as
    the category::

        sage: C.object([z, 1])
        Traceback (most recent call last):
        ...
        ValueError: constant coefficient must equal that of the category

    It is also possible to create a random object in the category. The
    input is the desired rank::

        sage: rho = C.random_object(2)
        sage: rho  # random
        Drinfeld module defined by T |--> (7*z^3 + 7*z^2 + 10*z + 2)*t^2 + (9*z^3 + 5*z^2 + 2*z + 7)*t + z^3 + 7*z^2 + 6*z + 10
        sage: rho.rank() == 2
        True
        sage: rho.category() is C
        True

    TESTS::

        sage: Fq = GF(11)
        sage: A.<T> = Fq[]
        sage: K.<z> = Fq.extension(4)
        sage: from sage.categories.drinfeld_modules import DrinfeldModules
        sage: base = Hom(A, K)(0)
        sage: C = DrinfeldModules(base)
        Traceback (most recent call last):
        ...
        TypeError: base field must be a ring extension

    ::

        sage: C.base().defining_morphism() == C.base_morphism()
        True

    ::

        sage: base = Hom(A, A)(1)
        sage: C = DrinfeldModules(base)
        Traceback (most recent call last):
        ...
        TypeError: base field must be a ring extension

    ::

        sage: base = 'I hate Rostropovitch'
        sage: C = DrinfeldModules(base)  # known bug (blankline)
        <BLANKLINE>
        Traceback (most recent call last):
        ...
        TypeError: input must be a ring morphism

    ::

        sage: ZZT.<T> = ZZ[]
        sage: base = Hom(ZZT, K)(1)
        sage: C = DrinfeldModules(base)  # known bug (blankline)
        <BLANKLINE>
        Traceback (most recent call last):
        ...
        TypeError: function ring base must be a finite field
    """

    def __init__(self, base_field, name='t'):
        r"""
        Initialize `self`.

        INPUT:

        - ``base_field`` -- the base field, which is a ring extension
          over a base

        - ``name`` (default: ``'t'``) -- the name of the Ore polynomial
          variable

        TESTS::

            sage: Fq = GF(11)
            sage: A.<T> = Fq[]
            sage: K.<z> = Fq.extension(4)
            sage: p_root = z^3 + 7*z^2 + 6*z + 10
            sage: phi = DrinfeldModule(A, [p_root, 0, 0, 1])
            sage: C = phi.category()
            sage: ore_polring.<t> = OrePolynomialRing(phi.base(), phi.base().frobenius_endomorphism())
            sage: C._ore_polring is ore_polring
            True
            sage: i = phi.base().coerce_map_from(K)
            sage: base_morphism = Hom(A, K)(p_root)
            sage: C.base() == K.over(base_morphism)
            True
            sage: C._base_morphism == i * base_morphism
            True
            sage: C._function_ring is A
            True
            sage: C._constant_coefficient == base_morphism(T)
            True
            sage: C._characteristic(C._constant_coefficient)
            0
        """
        # Check input is a ring extension
        if not isinstance(base_field, RingExtension_generic):
            raise TypeError('base field must be a ring extension')
        base_morphism = base_field.defining_morphism()
        self._base_morphism = base_morphism
        # Check input is a field
        if not base_field.is_field():
            raise TypeError('input must be a field')
        self._base_field = base_field
        self._function_ring = base_morphism.domain()
        # Check domain of base morphism is Fq[T]
        function_ring = self._function_ring
        if not isinstance(function_ring, PolynomialRing_general):
            raise NotImplementedError('function ring must be a polynomial '
                                      'ring')
        function_ring_base = function_ring.base_ring()
        if not function_ring_base.is_field() \
                or not function_ring_base.is_finite():
            raise TypeError('function ring base must be a finite field')
        # Shortcuts
        Fq = function_ring_base
        A = function_ring
        T = A.gen()
        K = base_field  # A ring extension
        # Build K{t}
        d = log(Fq.cardinality(), Fq.characteristic())
        tau = K.frobenius_endomorphism(d)
        self._ore_polring = OrePolynomialRing(K, tau, names=name,
                                              polcast=False)
        # Create constant coefficient
        self._constant_coefficient = base_morphism(T)
        # Create characteristic
        self._characteristic = None
        if K.is_finite():
            self._characteristic = A(K.over(Fq)(base_morphism(T)).minpoly())
        else:
            try:
                if base_morphism.is_injective():
                    self._characteristic = Integer(0)
            except NotImplementedError:
                pass
        # Create base over constants field
        i = A.coerce_map_from(Fq)
        Fq_to_K = self._base_morphism * i
        self._base_over_constants_field = base_field.over(Fq_to_K)
        super().__init__(base=base_field)

    def _latex_(self):
        r"""
        Return a latex representation of the category.

        OUTPUT: a string

        EXAMPLES::

            sage: Fq = GF(11)
            sage: A.<T> = Fq[]
            sage: K.<z> = Fq.extension(4)
            sage: p_root = z^3 + 7*z^2 + 6*z + 10
            sage: phi = DrinfeldModule(A, [p_root, 0, 0, 1])
            sage: C = phi.category()
            sage: latex(C)
            \text{Category{ }of{ }Drinfeld{ }modules{ }over{ }\Bold{F}_{11^{4}}
        """
        return f'\\text{{Category{{ }}of{{ }}Drinfeld{{ }}modules{{ }}' \
               f'over{{ }}{latex(self._base_field)}'

    def _repr_(self):
        r"""
        Return a string representation of the category.

        OUTPUT: a string

        EXAMPLES::

            sage: Fq = GF(11)
            sage: A.<T> = Fq[]
            sage: K.<z> = Fq.extension(4)
            sage: p_root = z^3 + 7*z^2 + 6*z + 10
            sage: phi = DrinfeldModule(A, [p_root, 0, 0, 1])
            sage: C = phi.category()
            sage: C
            Category of Drinfeld modules over Finite Field in z of size 11^4 over its base
        """
        return f'Category of Drinfeld modules over {self._base_field}'

    def Homsets(self):
        r"""
        Return the category of homsets.

        EXAMPLES::

            sage: Fq = GF(11)
            sage: A.<T> = Fq[]
            sage: K.<z> = Fq.extension(4)
            sage: p_root = z^3 + 7*z^2 + 6*z + 10
            sage: phi = DrinfeldModule(A, [p_root, 0, 0, 1])
            sage: C = phi.category()

            sage: from sage.categories.homsets import Homsets
            sage: C.Homsets() is Homsets()
            True
        """
        return Homsets()

    def Endsets(self):
        r"""
        Return the category of endsets.

        EXAMPLES::

            sage: Fq = GF(11)
            sage: A.<T> = Fq[]
            sage: K.<z> = Fq.extension(4)
            sage: p_root = z^3 + 7*z^2 + 6*z + 10
            sage: phi = DrinfeldModule(A, [p_root, 0, 0, 1])
            sage: C = phi.category()

            sage: from sage.categories.homsets import Homsets
            sage: C.Endsets() is Homsets().Endsets()
            True
        """
        return Homsets().Endsets()

    def base_morphism(self):
        r"""
        Return the base morphism of the category.

        EXAMPLES::

            sage: Fq = GF(11)
            sage: A.<T> = Fq[]
            sage: K.<z> = Fq.extension(4)
            sage: p_root = z^3 + 7*z^2 + 6*z + 10
            sage: phi = DrinfeldModule(A, [p_root, 0, 0, 1])
            sage: C = phi.category()
            sage: C.base_morphism()
            Ring morphism:
              From: Univariate Polynomial Ring in T over Finite Field of size 11
              To:   Finite Field in z of size 11^4 over its base
              Defn: T |--> z^3 + 7*z^2 + 6*z + 10

            sage: C.constant_coefficient() == C.base_morphism()(T)
            True
        """
        return self._base_morphism

    def base_over_constants_field(self):
        r"""
        Return the base field, seen as an extension over the constants
        field `\mathbb{F}_q`.

        EXAMPLES::

            sage: Fq = GF(11)
            sage: A.<T> = Fq[]
            sage: K.<z> = Fq.extension(4)
            sage: p_root = z^3 + 7*z^2 + 6*z + 10
            sage: phi = DrinfeldModule(A, [p_root, 0, 0, 1])
            sage: C = phi.category()
            sage: C.base_over_constants_field()
            Field in z with defining polynomial x^4 + 8*x^2 + 10*x + 2 over its base
        """
        return self._base_over_constants_field

    def characteristic(self):
        r"""
        Return the function ring-characteristic.

        EXAMPLES::

            sage: Fq = GF(11)
            sage: A.<T> = Fq[]
            sage: K.<z> = Fq.extension(4)
            sage: p_root = z^3 + 7*z^2 + 6*z + 10
            sage: phi = DrinfeldModule(A, [p_root, 0, 0, 1])
            sage: C = phi.category()
            sage: C.characteristic()
            T^2 + 7*T + 2

        ::

            sage: psi = DrinfeldModule(A, [Frac(A).gen(), 1])
            sage: C = psi.category()
            sage: C.characteristic()
            0
        """
        if self._characteristic is None:
            raise NotImplementedError('function ring characteristic not ' \
                                      'implemented in this case')
        return self._characteristic

    def constant_coefficient(self):
        r"""
        Return the constant coefficient of the category.

        EXAMPLES::

            sage: Fq = GF(11)
            sage: A.<T> = Fq[]
            sage: K.<z> = Fq.extension(4)
            sage: p_root = z^3 + 7*z^2 + 6*z + 10
            sage: phi = DrinfeldModule(A, [p_root, 0, 0, 1])
            sage: C = phi.category()
            sage: C.constant_coefficient()
            z^3 + 7*z^2 + 6*z + 10
            sage: C.constant_coefficient() == C.base()(T)
            True
        """
        return self._constant_coefficient

    def function_ring(self):
        r"""
        Return the function ring of the category.

        EXAMPLES::

            sage: Fq = GF(11)
            sage: A.<T> = Fq[]
            sage: K.<z> = Fq.extension(4)
            sage: p_root = z^3 + 7*z^2 + 6*z + 10
            sage: phi = DrinfeldModule(A, [p_root, 0, 0, 1])
            sage: C = phi.category()
            sage: C.function_ring()
            Univariate Polynomial Ring in T over Finite Field of size 11
            sage: C.function_ring() is A
            True
        """
        return self._function_ring

    def object(self, gen):
        r"""
        Return a Drinfeld module object in the category whose generator
        is the input.

        INPUT:

        - ``gen`` -- the generator of the Drinfeld module, given as an Ore
          polynomial or a list of coefficients

        EXAMPLES::

            sage: Fq = GF(11)
            sage: A.<T> = Fq[]
            sage: K.<z> = Fq.extension(4)
            sage: p_root = z^3 + 7*z^2 + 6*z + 10
            sage: psi = DrinfeldModule(A, [p_root, 1])
            sage: C = psi.category()

            sage: phi = C.object([p_root, 0, 1])
            sage: phi
            Drinfeld module defined by T |--> t^2 + z^3 + 7*z^2 + 6*z + 10
            sage: t = phi.ore_polring().gen()
            sage: C.object(t^2 + z^3 + 7*z^2 + 6*z + 10) is phi
            True
        """
        from sage.rings.function_field.drinfeld_modules.drinfeld_module import DrinfeldModule
        # If gen is not in the Ore polring, an exception is raised
        gen = self._ore_polring(gen)
        T = self._function_ring.gen()
        if gen[0] != self._base_morphism(T):
            raise ValueError('constant coefficient must equal that of the ' \
                             'category')
        return DrinfeldModule(self._function_ring, gen)

    def ore_polring(self):
        r"""
        Return the Ore polynomial ring of the category

        EXAMPLES::

            sage: Fq = GF(11)
            sage: A.<T> = Fq[]
            sage: K.<z> = Fq.extension(4)
            sage: p_root = z^3 + 7*z^2 + 6*z + 10
            sage: phi = DrinfeldModule(A, [p_root, 0, 0, 1])
            sage: C = phi.category()
            sage: C.ore_polring()
            Ore Polynomial Ring in t over Finite Field in z of size 11^4 over its base twisted by Frob
        """
        return self._ore_polring

    def random_object(self, rank):
        r"""
        Return a random Drinfeld module in the category with given rank.

        INPUT:

        - ``rank`` -- an integer, the rank of the Drinfeld module

        EXAMPLES::

            sage: Fq = GF(11)
            sage: A.<T> = Fq[]
            sage: K.<z> = Fq.extension(4)
            sage: p_root = z^3 + 7*z^2 + 6*z + 10
            sage: phi = DrinfeldModule(A, [p_root, 0, 0, 1])
            sage: C = phi.category()

            sage: psi = C.random_object(3)  # random
            Drinfeld module defined by T |--> (6*z^3 + 4*z^2 + 10*z + 9)*t^3 + (4*z^3 + 8*z^2 + 8*z)*t^2 + (10*z^3 + 3*z^2 + 6*z)*t + z^3 + 7*z^2 + 6*z + 10
            sage: psi.rank() == 3
            True
        """
        if not isinstance(rank, Integer):
            raise TypeError('rank must be a positive integer')
        if rank <= 0:
            raise ValueError('rank must be a positive integer')

        K = self._base_field
        coeffs = [self._constant_coefficient]
        for _ in range(rank-1):
            coeffs.append(K.random_element())
        dom_coeff = 0
        while dom_coeff == 0:
            dom_coeff = K.random_element()
        coeffs.append(dom_coeff)

        return self.object(coeffs)

    def super_categories(self):
        """
        EXAMPLES::

            sage: Fq = GF(11)
            sage: A.<T> = Fq[]
            sage: K.<z> = Fq.extension(4)
            sage: p_root = z^3 + 7*z^2 + 6*z + 10
            sage: phi = DrinfeldModule(A, [p_root, 0, 0, 1])
            sage: C = phi.category()
            sage: C.super_categories()
            [Category of objects]
        """
        return [Objects()]

    class ParentMethods:

        def base(self):
            r"""
            Return the base field of this Drinfeld module, viewed as
            an algebra over the function ring.

            This is an instance of the class
            :class:`sage.rings.ring_extension.RingExtension`.

            EXAMPLES::

                sage: Fq = GF(25)
                sage: A.<T> = Fq[]
                sage: K.<z12> = Fq.extension(6)
                sage: p_root = 2*z12^11 + 2*z12^10 + z12^9 + 3*z12^8 + z12^7 + 2*z12^5 + 2*z12^4 + 3*z12^3 + z12^2 + 2*z12
                sage: phi = DrinfeldModule(A, [p_root, z12^3, z12^5])
                sage: phi.base()
                Finite Field in z12 of size 5^12 over its base

            The base can be infinite::

                sage: sigma = DrinfeldModule(A, [Frac(A).gen(), 1])
                sage: sigma.base()
                Fraction Field of Univariate Polynomial Ring in T over Finite Field in z2 of size 5^2 over its base
            """
            return self.category().base()

        def base_morphism(self):
            r"""
            Return the base morphism of this Drinfeld module.

            EXAMPLES::

                sage: Fq = GF(25)
                sage: A.<T> = Fq[]
                sage: K.<z12> = Fq.extension(6)
                sage: p_root = 2*z12^11 + 2*z12^10 + z12^9 + 3*z12^8 + z12^7 + 2*z12^5 + 2*z12^4 + 3*z12^3 + z12^2 + 2*z12
                sage: phi = DrinfeldModule(A, [p_root, z12^3, z12^5])
                sage: phi.base_morphism()
                Ring morphism:
                  From: Univariate Polynomial Ring in T over Finite Field in z2 of size 5^2
                  To:   Finite Field in z12 of size 5^12 over its base
                  Defn: T |--> 2*z12^11 + 2*z12^10 + z12^9 + 3*z12^8 + z12^7 + 2*z12^5 + 2*z12^4 + 3*z12^3 + z12^2 + 2*z12

            The base field can be infinite::

                sage: sigma = DrinfeldModule(A, [Frac(A).gen(), 1])
                sage: sigma.base_morphism()
                Ring morphism:
                  From: Univariate Polynomial Ring in T over Finite Field in z2 of size 5^2
                  To:   Fraction Field of Univariate Polynomial Ring in T over Finite Field in z2 of size 5^2 over its base
                  Defn: T |--> T
            """
            return self.category().base_morphism()

        def base_over_constants_field(self):
            r"""
            Return the base field, seen as an extension over the constants
            field `\mathbb{F}_q`.

            This is an instance of the class
            :class:`sage.rings.ring_extension.RingExtension`.

            EXAMPLES::

                sage: Fq = GF(25)
                sage: A.<T> = Fq[]
                sage: K.<z12> = Fq.extension(6)
                sage: p_root = 2*z12^11 + 2*z12^10 + z12^9 + 3*z12^8 + z12^7 + 2*z12^5 + 2*z12^4 + 3*z12^3 + z12^2 + 2*z12
                sage: phi = DrinfeldModule(A, [p_root, z12^3, z12^5])
                sage: phi.base_over_constants_field()
                Field in z12 with defining polynomial x^6 + (4*z2 + 3)*x^5 + x^4 + (3*z2 + 1)*x^3 + x^2 + (4*z2 + 1)*x + z2 over its base
            """
            return self.category().base_over_constants_field()

        def characteristic(self):
            r"""
            Return the function ring-characteristic.

            EXAMPLES::

                sage: Fq = GF(25)
                sage: A.<T> = Fq[]
                sage: K.<z12> = Fq.extension(6)
                sage: p_root = 2*z12^11 + 2*z12^10 + z12^9 + 3*z12^8 + z12^7 + 2*z12^5 + 2*z12^4 + 3*z12^3 + z12^2 + 2*z12
                sage: phi = DrinfeldModule(A, [p_root, z12^3, z12^5])
                sage: phi.characteristic()
                T^2 + (4*z2 + 2)*T + 2
                sage: phi.base_morphism()(phi.characteristic())
                0

            ::

                sage: B.<Y> = Fq[]
                sage: L = Frac(B)
                sage: psi = DrinfeldModule(A, [L(1), 0, 0, L(1)])
                sage: psi.characteristic()
                Traceback (most recent call last):
                ...
                NotImplementedError: function ring characteristic not implemented in this case
            """
            return self.category().characteristic()

        def function_ring(self):
            r"""
            Return the function ring of this Drinfeld module.

            EXAMPLES::

                sage: Fq = GF(25)
                sage: A.<T> = Fq[]
                sage: K.<z12> = Fq.extension(6)
                sage: p_root = 2*z12^11 + 2*z12^10 + z12^9 + 3*z12^8 + z12^7 + 2*z12^5 + 2*z12^4 + 3*z12^3 + z12^2 + 2*z12
                sage: phi = DrinfeldModule(A, [p_root, z12^3, z12^5])
                sage: phi.function_ring()
                Univariate Polynomial Ring in T over Finite Field in z2 of size 5^2
                sage: phi.function_ring() is A
                True
            """
            return self.category().function_ring()

        def constant_coefficient(self):
            r"""
            Return the constant coefficient of the generator
            of this Drinfeld module.

            OUTPUT: an element in the base field

            EXAMPLES::

                sage: Fq = GF(25)
                sage: A.<T> = Fq[]
                sage: K.<z12> = Fq.extension(6)
                sage: p_root = 2*z12^11 + 2*z12^10 + z12^9 + 3*z12^8 + z12^7 + 2*z12^5 + 2*z12^4 + 3*z12^3 + z12^2 + 2*z12
                sage: phi = DrinfeldModule(A, [p_root, z12^3, z12^5])
                sage: phi.constant_coefficient() == p_root
                True

            Let `\mathbb{F}_q[T]` be the function ring, and let `\gamma` be
            the base of the Drinfeld module. The constant coefficient is
            `\gamma(T)`::

                sage: C = phi.category()
                sage: base = C.base()
                sage: base(T) == phi.constant_coefficient()
                True

            Naturally, two Drinfeld modules in the same category have the
            same constant coefficient::

                sage: t = phi.ore_polring().gen()
                sage: psi = C.object(phi.constant_coefficient() + t^3)
                sage: psi
                Drinfeld module defined by T |--> t^3 + 2*z12^11 + 2*z12^10 + z12^9 + 3*z12^8 + z12^7 + 2*z12^5 + 2*z12^4 + 3*z12^3 + z12^2 + 2*z12

            Reciprocally, it is impossible to create two Drinfeld modules in
            this category if they do not share the same constant
            coefficient::

                sage: rho = C.object(phi.constant_coefficient() + 1 + t^3)
                Traceback (most recent call last):
                ...
                ValueError: constant coefficient must equal that of the category
            """
            return self.category().constant_coefficient()

        def ore_polring(self):
            r"""
            Return the Ore polynomial ring of this Drinfeld module.

            EXAMPLES::

                sage: Fq = GF(25)
                sage: A.<T> = Fq[]
                sage: K.<z12> = Fq.extension(6)
                sage: p_root = 2*z12^11 + 2*z12^10 + z12^9 + 3*z12^8 + z12^7 + 2*z12^5 + 2*z12^4 + 3*z12^3 + z12^2 + 2*z12
                sage: phi = DrinfeldModule(A, [p_root, z12^3, z12^5])
                sage: S = phi.ore_polring()
                sage: S
                Ore Polynomial Ring in t over Finite Field in z12 of size 5^12 over its base twisted by Frob^2

            The Ore polynomial ring can also be retrieved from the category
            of the Drinfeld module::

                sage: S is phi.category().ore_polring()
                True

            The generator of the Drinfeld module is in the Ore polynomial
            ring::

                sage: phi(T) in S
                True
            """
            return self.category().ore_polring()



r"""
Filtered Algebras With Basis

A filtered algebra with basis over a commutative ring `R`
is a filtered algebra over `R` endowed with the structure
of a filtered module with basis (with the same underlying
filtered-module structure). See
:class:`~sage.categories.filtered_algebras.FilteredAlgebras` and
:class:`~sage.categories.filtered_modules_with_basis.FilteredModulesWithBasis`
for these two notions.
"""
#*****************************************************************************
#  Copyright (C) 2014 Travis Scrimshaw <tscrim at ucdavis.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#******************************************************************************

from sage.categories.filtered_modules import FilteredModulesCategory

class FilteredAlgebrasWithBasis(FilteredModulesCategory):
    """
    The category of filtered algebras with a distinguished
    homogeneous basis.

    A filtered algebra with basis over a commutative ring `R`
    is a filtered algebra over `R` endowed with the structure
    of a filtered module with basis (with the same underlying
    filtered-module structure). See
    :class:`~sage.categories.filtered_algebras.FilteredAlgebras` and
    :class:`~sage.categories.filtered_modules_with_basis.FilteredModulesWithBasis`
    for these two notions.

    EXAMPLES::

        sage: C = AlgebrasWithBasis(ZZ).Filtered(); C
        Category of filtered algebras with basis over Integer Ring
        sage: sorted(C.super_categories(), key=str)
        [Category of algebras with basis over Integer Ring,
         Category of filtered algebras over Integer Ring,
         Category of filtered modules with basis over Integer Ring]

    TESTS::

        sage: TestSuite(C).run()
    """
    class ParentMethods:
        def graded_algebra(self):
            r"""
            Return the associated graded algebra to ``self``.

            See :class:`~sage.algebras.associated_graded.AssociatedGradedAlgebra`
            for the definition and the properties of this.

            If the filtered algebra ``self`` with basis is called `A`,
            then this method returns `\operatorname{gr} A`. The method
            :meth:`to_graded_conversion` returns the canonical
            `R`-module isomorphism `A \to \operatorname{gr} A` induced
            by the basis of `A`, and the method
            :meth:`from_graded_conversion` returns the inverse of this
            isomorphism. The method :meth:`projection` projects
            elements of `A` onto `\operatorname{gr} A` according to
            their place in the filtration on `A`.

            .. WARNING::

                When not overridden, this method returns the default
                implementation of an associated graded algebra --
                namely, ``AssociatedGradedAlgebra(self)``, where
                ``AssociatedGradedAlgebra`` is
                :class:`~sage.algebras.associated_graded.AssociatedGradedAlgebra`.
                But many instances of :class:`FilteredAlgebrasWithBasis`
                override this method, as the associated graded algebra
                often is (isomorphic) to a simpler object (for instance,
                the associated graded algebra of a graded algebra can be
                identified with the graded algebra itself). Generic code
                that uses associated graded algebras (such as the code
                of the :meth:`induced_graded_map` method below) should
                make sure to only communicate with them via the
                :meth:`to_graded_conversion`,
                :meth:`from_graded_conversion`, and
                :meth:`projection` methods (in particular,
                do not expect there to be a conversion from ``self``
                to ``self.graded_algebra()``; this currently does not
                work for Clifford algebras). Similarly, when
                overriding :meth:`graded_algebra`, make sure to
                accordingly redefine these three methods, unless their
                definitions below still apply to your case (this will
                happen whenever the basis of your :meth:`graded_algebra`
                has the same indexing set as ``self``, and the partition
                of this indexing set according to degree is the same as
                for ``self``).

            .. TODO::

                Maybe the thing about the conversion from ``self``
                to ``self.graded_algebra()`` on the Clifford at least
                could be made to work? (I would still warn the user
                against ASSUMING that it must work -- as there is
                probably no way to guarantee it in all cases, and
                we shouldn't require users to mess with
                element constructors.)

            EXAMPLES::

                sage: A = AlgebrasWithBasis(ZZ).Filtered().example()
                sage: A.graded_algebra()
                Graded Algebra of An example of a filtered algebra with basis:
                 the universal enveloping algebra of
                 Lie algebra of RR^3 with cross product over Integer Ring
            """
            from sage.algebras.associated_graded import AssociatedGradedAlgebra
            return AssociatedGradedAlgebra(self)

        # Maps

        def to_graded_conversion(self):
            r"""
            Return the canonical `R`-module isomorphism
            `A \to \operatorname{gr} A` induced by the basis of `A`
            (where `A = ` ``self``).

            This is an isomorphism of `R`-modules, not of algebras. See
            the class documentation :class:`AssociatedGradedAlgebra`.

            .. SEEALSO::

                :meth:`from_graded_conversion`

            EXAMPLES::

                sage: A = Algebras(QQ).WithBasis().Filtered().example()
                sage: p = A.an_element() + A.algebra_generators()['x'] + 2; p
                U['x']^2*U['y']^2*U['z']^3 + 3*U['x'] + 3*U['y'] + 3
                sage: q = A.to_graded_conversion()(p); q
                bar(U['x']^2*U['y']^2*U['z']^3) + 3*bar(U['x'])
                 + 3*bar(U['y']) + 3*bar(1)
                sage: q.parent() is A.graded_algebra()
                True
            """
            base_one = self.base_ring().one()
            return self.module_morphism(diagonal=lambda x: base_one,
                                        codomain=self.graded_algebra())

        def from_graded_conversion(self):
            r"""
            Return the inverse of the canonical `R`-module isomorphism
            `A \to \operatorname{gr} A` induced by the basis of `A`
            (where `A = ` ``self``). This inverse is an isomorphism
            `\operatorname{gr} A \to A`.

            This is an isomorphism of `R`-modules, not of algebras. See
            the class documentation :class:`AssociatedGradedAlgebra`.

            .. SEEALSO::

                :meth:`to_graded_conversion`

            EXAMPLES::

                sage: A = Algebras(QQ).WithBasis().Filtered().example()
                sage: p = A.an_element() + A.algebra_generators()['x'] + 2; p
                U['x']^2*U['y']^2*U['z']^3 + 3*U['x'] + 3*U['y'] + 3
                sage: q = A.to_graded_conversion()(p)
                sage: A.from_graded_conversion()(q) == p
                True
                sage: q.parent() is A.graded_algebra()
                True
            """
            base_one = self.base_ring().one()
            return self.graded_algebra().module_morphism(diagonal=lambda x: base_one,
                                                         codomain=self)

        def projection(self, i):
            r"""
            Return the `i`-th projection `p_i : F_i \to G_i` (in the
            notations of the class documentation
            :class:`AssociatedGradedAlgebra`, where `A = ` ``self``).

            This method actually does not return the map `p_i` itself,
            but an extension of `p_i` to the whole `R`-module `A`.
            This extension is the composition of the `R`-module
            isomorphism `A \to \operatorname{gr} A` with the canonical
            projection of the graded `R`-module `\operatorname{gr} A`
            onto its `i`-th graded component `G_i`. The codomain of
            this map is `\operatorname{gr} A`, although its actual
            image is `G_i`. The map `p_i` is obtained from this map
            by restricting its domain to `F_i` and its image to `G_i`.

            EXAMPLES::

                sage: A = Algebras(QQ).WithBasis().Filtered().example()
                sage: p = A.an_element() + A.algebra_generators()['x'] + 2; p
                U['x']^2*U['y']^2*U['z']^3 + 3*U['x'] + 3*U['y'] + 3
                sage: q = A.projection(7)(p); q
                bar(U['x']^2*U['y']^2*U['z']^3)
                sage: q.parent() is A.graded_algebra()
                True
                sage: A.projection(8)(p)
                0
            """
            base_zero = self.base_ring().zero()
            base_one = self.base_ring().one()
            grA = self.graded_algebra()
            proj = lambda x: (base_one if self.degree_on_basis(x) == i
                              else base_zero)
            return self.module_morphism(diagonal=proj, codomain=grA)

        def induced_graded_map(self, other, f):
            r"""
            Return the graded linear map between the associated graded
            algebras of ``self`` and ``other`` canonically induced by
            the filtration-preserving map ``f : self -> other``.

            Let `A` and `B` be two filtered algebras with basis, and let
            `(F_i)_{i \in I}` and `(G_i)_{i \in I}` be their
            filtrations. Let `f : A \to B` be a linear map which
            preserves the filtration (i.e., satisfies `f(F_i) \subseteq
            G_i` for all `i \in I`). Then, there is a canonically
            defined graded linear map
            `\operatorname{gr} f : \operatorname{gr} A \to
            \operatorname{gr} B` which satisfies

            .. MATH::

                (\operatorname{gr} f) (p_i(a)) = p_i(f(a))
                \qquad \text{for all } i \in I \text{ and } a \in F_i ,

            where the `p_i` on the left hand side is the canonical
            projection from `F_i` onto the `i`-th graded component
            of `\operatorname{gr} A`, while the `p_i` on the right
            hand side is the canonical projection from `G_i` onto
            the `i`-th graded component of `\operatorname{gr} B`.

            INPUT:

            - ``other`` -- a filtered algebra with basis

            - ``f`` -- a filtration-preserving linear map from ``self``
              to ``other`` (can be given as a morphism or as a function)

            OUTPUT:

            The graded linear map `\operatorname{gr} f`.

            EXAMPLES:

            **Example 1.**

            We start with the universal enveloping algebra of the
            Lie algebra `\RR^3` (with the cross product serving as
            Lie bracket)::

                sage: A = AlgebrasWithBasis(QQ).Filtered().example(); A
                An example of a filtered algebra with basis: the
                 universal enveloping algebra of Lie algebra of RR^3
                 with cross product over Rational Field
                sage: M = A.indices(); M
                Free abelian monoid indexed by {'x', 'y', 'z'}
                sage: x,y,z = [A.basis()[M.gens()[i]] for i in "xyz"]

            Let us define a stupid filtered map from ``A`` to
            itself::

                sage: def map_on_basis(m):
                ....:     d = m.dict()
                ....:     i = d.get('x', 0); j = d.get('y', 0); k = d.get('z', 0)
                ....:     g = (y ** (i+j)) * (z ** k)
                ....:     if i > 0:
                ....:         g += i * (x ** (i-1)) * (y ** j) * (z ** k)
                ....:     return g
                sage: f = A.module_morphism(on_basis=map_on_basis,
                ....:                       codomain=A)
                sage: f(x)
                U['y'] + 1
                sage: f(x*y*z)
                U['y']^2*U['z'] + U['y']*U['z']
                sage: f(x*x*y*z)
                U['y']^3*U['z'] + 2*U['x']*U['y']*U['z']
                sage: f(A.one())
                1
                sage: f(y*z)
                U['y']*U['z']

            (There is nothing here that is peculiar to this
            universal enveloping algebra; we are only using its
            module structure, and we could just as well be using
            a polynomial algebra in its stead.)

            We now compute `\operatorname{gr} f` ::

                sage: grA = A.graded_algebra(); grA
                Graded Algebra of An example of a filtered algebra with
                 basis: the universal enveloping algebra of Lie algebra
                 of RR^3 with cross product over Rational Field
                sage: xx, yy, zz = [A.to_graded_conversion()(i) for i in [x, y, z]]
                sage: xx+yy*zz
                bar(U['y']*U['z']) + bar(U['x'])
                sage: grf = A.induced_graded_map(A, f); grf
                Generic endomorphism of Graded Algebra of An example
                 of a filtered algebra with basis: the universal
                 enveloping algebra of Lie algebra of RR^3 with cross
                 product over Rational Field
                sage: grf(xx)
                bar(U['y'])
                sage: grf(xx*yy*zz)
                bar(U['y']^2*U['z'])
                sage: grf(xx*xx*yy*zz)
                bar(U['y']^3*U['z'])
                sage: grf(grA.one())
                bar(1)
                sage: grf(yy*zz)
                bar(U['y']*U['z'])
                sage: grf(yy*zz-2*yy)
                bar(U['y']*U['z']) - 2*bar(U['y'])

            **Example 2.**

            We shall now construct `\operatorname{gr} f` for a
            different map `f` out of the same ``A``; the new map
            `f` will lead into a graded algebra already, namely into
            the algebra of symmetric functions::

                sage: h = SymmetricFunctions(QQ).h()
                sage: def map_on_basis(m):  # redefining map_on_basis
                ....:     d = m.dict()
                ....:     i = d.get('x', 0); j = d.get('y', 0); k = d.get('z', 0)
                ....:     g = (h[1] ** i) * (h[2] ** (j // 2) * (h[3] ** (k // 3)))
                ....:     g += i * (h[1] ** (i+j+k))
                ....:     return g
                sage: f = A.module_morphism(on_basis=map_on_basis,
                ....:                       codomain=h)  # redefining f
                sage: f(x)
                2*h[1]
                sage: f(y)
                h[]
                sage: f(z)
                h[]
                sage: f(y**2)
                h[2]
                sage: f(x**2)
                3*h[1, 1]
                sage: f(x*y*z)
                h[1] + h[1, 1, 1]
                sage: f(x*x*y*y*z)
                2*h[1, 1, 1, 1, 1] + h[2, 1, 1]
                sage: f(A.one())
                h[]

            The algebra ``h`` of symmetric functions in the `h`-basis
            is already graded, so its associated graded algebra is
            implemented as itself::

                sage: grh = h.graded_algebra(); grh is h
                True
                sage: grf = A.induced_graded_map(h, f); grf
                Generic morphism:
                  From: Graded Algebra of An example of a filtered
                   algebra with basis: the universal enveloping
                   algebra of Lie algebra of RR^3 with cross
                   product over Rational Field
                  To:   Symmetric Functions over Rational Field
                   in the homogeneous basis
                sage: grf(xx)
                2*h[1]
                sage: grf(yy)
                0
                sage: grf(zz)
                0
                sage: grf(yy**2)
                h[2]
                sage: grf(xx**2)
                3*h[1, 1]
                sage: grf(xx*yy*zz)
                h[1, 1, 1]
                sage: grf(xx*xx*yy*yy*zz)
                2*h[1, 1, 1, 1, 1]
                sage: grf(grA.one())
                h[]

            **Example 3.**

            After having had a graded algebra as the codomain, let us try to
            have one as the domain instead. Our new ``f`` will go from ``h``
            to ``A``::

                sage: def map_on_basis(lam):  # redefining map_on_basis
                ....:     return x ** (sum(lam)) + y ** (len(lam))
                sage: f = h.module_morphism(on_basis=map_on_basis,
                ....:                       codomain=A)  # redefining f
                sage: f(h[1])
                U['x'] + U['y']
                sage: f(h[2])
                U['x']^2 + U['y']
                sage: f(h[1, 1])
                U['x']^2 + U['y']^2
                sage: f(h[2, 2])
                U['x']^4 + U['y']^2
                sage: f(h[3, 2, 1])
                U['x']^6 + U['y']^3
                sage: f(h.one())
                2
                sage: grf = h.induced_graded_map(A, f); grf
                Generic morphism:
                  From: Symmetric Functions over Rational Field
                   in the homogeneous basis
                  To:   Graded Algebra of An example of a filtered
                   algebra with basis: the universal enveloping
                   algebra of Lie algebra of RR^3 with cross
                   product over Rational Field
                sage: grf(h[1])
                bar(U['x']) + bar(U['y'])
                sage: grf(h[2])
                bar(U['x']^2)
                sage: grf(h[1, 1])
                bar(U['x']^2) + bar(U['y']^2)
                sage: grf(h[2, 2])
                bar(U['x']^4)
                sage: grf(h[3, 2, 1])
                bar(U['x']^6)
                sage: grf(h.one())
                2*bar(1)

            **Example 4.**

            The construct `\operatorname{gr} f` also makes sense when `f`
            is a filtration-preserving map between graded algebras. ::

                sage: def map_on_basis(lam):  # redefining map_on_basis
                ....:     return h[lam] + h[len(lam)]
                sage: f = h.module_morphism(on_basis=map_on_basis,
                ....:                       codomain=h)  # redefining f
                sage: f(h[1])
                2*h[1]
                sage: f(h[2])
                h[1] + h[2]
                sage: f(h[1, 1])
                h[1, 1] + h[2]
                sage: f(h[2, 1])
                h[2] + h[2, 1]
                sage: f(h.one())
                2*h[]
                sage: grf = h.induced_graded_map(h, f); grf
                Generic endomorphism of Symmetric Functions over Rational
                 Field in the homogeneous basis
                sage: grf(h[1])
                2*h[1]
                sage: grf(h[2])
                h[2]
                sage: grf(h[1, 1])
                h[1, 1] + h[2]
                sage: grf(h[2, 1])
                h[2, 1]
                sage: grf(h.one())
                2*h[]

            **Example 5.**

            For another example, let us compute `\operatorname{gr} f` for a
            map `f` between two Clifford algebras::

                sage: Q = QuadraticForm(ZZ, 2, [1,2,3])
                sage: B = CliffordAlgebra(Q, names=['u','v']); B
                The Clifford algebra of the Quadratic form in 2
                 variables over Integer Ring with coefficients:
                [ 1 2 ]
                [ * 3 ]
                sage: m = Matrix(ZZ, [[1, 2], [1, -1]])
                sage: f = B.lift_module_morphism(m, names=['x','y'])
                sage: A = f.domain(); A
                The Clifford algebra of the Quadratic form in 2
                 variables over Integer Ring with coefficients:
                [ 6 0 ]
                [ * 3 ]
                sage: x, y = A.gens()
                sage: f(x)
                u + v
                sage: f(y)
                2*u - v
                sage: f(x**2)
                6
                sage: f(x*y)
                -3*u*v + 3
                sage: grA = A.graded_algebra(); grA
                The exterior algebra of rank 2 over Integer Ring
                sage: A.to_graded_conversion()(x)
                x
                sage: A.to_graded_conversion()(y)
                y
                sage: A.to_graded_conversion()(x*y)
                x*y
                sage: u = A.to_graded_conversion()(x*y+1); u
                x*y + 1
                sage: A.from_graded_conversion()(u)
                x*y + 1
                sage: A.projection(2)(x*y+1)
                x*y
                sage: A.projection(1)(x+2*y-2)
                x + 2*y
                sage: grf = A.induced_graded_map(B, f); grf
                Generic morphism:
                  From: The exterior algebra of rank 2 over Integer Ring
                  To:   The exterior algebra of rank 2 over Integer Ring
                sage: grf(A.to_graded_conversion()(x))
                u + v
                sage: grf(A.to_graded_conversion()(y))
                2*u - v
                sage: grf(A.to_graded_conversion()(x**2))
                6
                sage: grf(A.to_graded_conversion()(x*y))
                -3*u*v
                sage: grf(grA.one())
                1
            """
            grA = self.graded_algebra()
            grB = other.graded_algebra()
            from sage.categories.graded_modules_with_basis import GradedModulesWithBasis
            cat = GradedModulesWithBasis(self.base_ring())
            from_gr = self.from_graded_conversion()

            def on_basis(m):
                i = grA.degree_on_basis(m)
                lifted_img_of_m = f(from_gr(grA.monomial(m)))
                return other.projection(i)(lifted_img_of_m)
            return grA.module_morphism(on_basis=on_basis,
                                       codomain=grB, category=cat)
            # If we could assume that the projection of the basis
            # element of ``self`` indexed by an index ``m`` is the
            # basis element of ``grA`` indexed by ``m``, then this
            # could go faster:
            #
            # def on_basis(m):
            #     i = grA.degree_on_basis(m)
            #     return grB.projection(i)(f(self.monomial(m)))
            # return grA.module_morphism(on_basis=on_basis,
            #                            codomain=grB, category=cat)
            #
            # But this assumption might come back to bite us in the
            # ass one day. What do you think?

    class ElementMethods:
        pass



# Subset of sage.categories.all that is made available by the sage-objects distribution

from sage.misc.lazy_import import lazy_import

# Resolve a circular import so that "import sage.categories.all" can succeed
# in initializing the category system.
import sage.structure.category_object    # imports sage.categories.category

# Small part of "from .basic import *":
from .objects import Objects
from .sets_cat import Sets, EmptySetError


from .category import Category

from .category_types import Elements

from .cartesian_product import cartesian_product

from .functor  import (ForgetfulFunctor,
                      IdentityFunctor)

from .homset   import (Hom, hom,
                      End, end,
                      Homset, HomsetWithBase)

from .morphism import Morphism

from .realizations import Realizations

from .sets_with_partial_maps import SetsWithPartialMaps



r"""
Monoid algebras
"""
# ****************************************************************************
#  Copyright (C) 2005      David Kohel <kohel@maths.usyd.edu>
#                          William Stein <wstein@math.ucsd.edu>
#                2008-2009 Nicolas M. Thiery <nthiery at users.sf.net>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# *****************************************************************************


def MonoidAlgebras(base_ring):
    """
    The category of monoid algebras over ``base_ring``.

    EXAMPLES::

        sage: C = MonoidAlgebras(QQ); C
        Category of monoid algebras over Rational Field
        sage: sorted(C.super_categories(), key=str)
        [Category of bialgebras with basis over Rational Field,
         Category of semigroup algebras over Rational Field,
         Category of unital magma algebras over Rational Field]

    This is just an alias for::

        sage: C is Monoids().Algebras(QQ)
        True

    TESTS::

        sage: TestSuite(MonoidAlgebras(ZZ)).run()
    """
    from sage.categories.monoids import Monoids
    return Monoids().Algebras(base_ring)



r"""
Bialgebras
"""
# ****************************************************************************
#  Copyright (C) 2008 Teresa Gomez-Diaz (CNRS) <Teresa.Gomez-Diaz@univ-mlv.fr>
#  Copyright (C) 2008-2009 Nicolas M. Thiery <nthiery at users.sf.net>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# *****************************************************************************

from sage.categories.category_types import Category_over_base_ring
from sage.categories.algebras import Algebras
from sage.categories.coalgebras import Coalgebras
from sage.categories.super_modules import SuperModulesCategory
from sage.misc.lazy_import import LazyImport


class Bialgebras(Category_over_base_ring):
    """
    The category of bialgebras

    EXAMPLES::

        sage: Bialgebras(ZZ)
        Category of bialgebras over Integer Ring
        sage: Bialgebras(ZZ).super_categories()
        [Category of algebras over Integer Ring, Category of coalgebras over Integer Ring]

    TESTS::

        sage: TestSuite(Bialgebras(ZZ)).run()
    """

    def super_categories(self):
        """
        EXAMPLES::

            sage: Bialgebras(QQ).super_categories()
            [Category of algebras over Rational Field, Category of coalgebras over Rational Field]
        """
        R = self.base_ring()
        return [Algebras(R), Coalgebras(R)]

    def additional_structure(self):
        r"""
        Return ``None``.

        Indeed, the category of bialgebras defines no additional
        structure: a morphism of coalgebras and of algebras between
        two bialgebras is a bialgebra morphism.

        .. SEEALSO:: :meth:`Category.additional_structure`

        .. TODO:: This category should be a :class:`CategoryWithAxiom`.

        EXAMPLES::

            sage: Bialgebras(QQ).additional_structure()
        """
        return None

    class ElementMethods:

        def is_primitive(self):
            """
            Return whether ``self`` is a primitive element.

            EXAMPLES::

                sage: s = SymmetricFunctions(QQ).schur()
                sage: s([5]).is_primitive()
                False
                sage: p = SymmetricFunctions(QQ).powersum()
                sage: p([5]).is_primitive()
                True
            """
            one = self.parent().one()
            return self.coproduct() == one.tensor(self) + self.tensor(one)

        def is_grouplike(self):
            """
            Return whether ``self`` is a grouplike element.

            EXAMPLES::

                sage: s = SymmetricFunctions(QQ).schur()
                sage: s([5]).is_grouplike()
                False
                sage: s([]).is_grouplike()
                True
            """
            return self.coproduct() == self.tensor(self)

    class Super(SuperModulesCategory):
        pass

    WithBasis = LazyImport('sage.categories.bialgebras_with_basis', 'BialgebrasWithBasis')
